# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Generators for the grid"""

from typing import Generic, Type, TypeVar

import numpy as np

from power_grid_model_ds._core.data_source.generator.arrays.line import LineGenerator
from power_grid_model_ds._core.data_source.generator.arrays.node import NodeGenerator
from power_grid_model_ds._core.data_source.generator.arrays.source import SourceGenerator
from power_grid_model_ds._core.data_source.generator.arrays.transformer import TransformerGenerator
from power_grid_model_ds._core.model.graphs.models.base import BaseGraphModel
from power_grid_model_ds._core.model.graphs.models.rustworkx import RustworkxGraphModel
from power_grid_model_ds._core.model.grids.base import Grid

# pylint: disable=too-few-public-methods,too-many-arguments,too-many-positional-arguments

T = TypeVar("T", bound=Grid)


class RadialGridGenerator(Generic[T]):
    """Generates a random but structurally correct radial grid with the given specifications"""

    def __init__(
        self,
        grid_class: Type[T],
        nr_nodes: int = 100,
        nr_sources: int = 2,
        nr_nops: int = 10,
        graph_model: type[BaseGraphModel] = RustworkxGraphModel,
    ):
        self.grid_class = grid_class
        self.graph_model = graph_model
        self.nr_nodes = nr_nodes
        self.nr_sources = nr_sources
        self.nr_nops = nr_nops

    def run(self, seed=None, create_10_3_kv_net: bool = False) -> T:
        """Run the generator to create a random radial grid.

        if a seed is provided, this will be used to set rng.
        """
        grid = self.grid_class.empty(graph_model=self.graph_model)

        # create nodeArray
        node_generator = NodeGenerator(grid=grid, seed=seed)

        nodes, _loads_low, loads_high = node_generator.run(amount=self.nr_nodes)
        grid.append(nodes)
        grid.append(loads_high)

        # create sourceArray
        source_generator = SourceGenerator(grid=grid, seed=seed)
        nodes, sources = source_generator.run(amount=self.nr_sources)
        grid.append(nodes)
        grid.append(sources)

        # create lineArray
        line_generator = LineGenerator(grid=grid, seed=seed)
        lines = line_generator.run(amount=self.nr_nops)
        grid.append(lines)

        if create_10_3_kv_net:
            # create 3kV nodes
            nodes, _loads_low, _loads_high = node_generator.run(amount=10, voltage_level=3_000)
            grid.append(nodes)
            grid.append(_loads_high)

            # create transformerArray
            transformer_generator = TransformerGenerator(grid=grid, seed=seed)
            transformers = transformer_generator.run(amount=2)
            grid.append(transformers)

            lines = line_generator.run(amount=0, number_of_routes=0)
            grid.append(lines[~np.isin(lines.id, grid.line.id)])

        return grid



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Generator for NodeArray"""

import numpy as np

from power_grid_model_ds._core.data_source.generator.arrays.base import BaseGenerator


class NodeGenerator(BaseGenerator):
    """Generator for node elements in the grid"""

    # pylint: disable=arguments-differ
    def run(self, amount: int, voltage_level: int = 10_500):
        """Generate nodes in a grid with two possible load scenarios"""
        node_array = self.grid.node.__class__.zeros(amount)
        node_array.id = 1 + self.grid.max_id + np.arange(amount)
        node_array.u_rated = voltage_level

        load_low_array = self.grid.sym_load.__class__.zeros(amount)
        load_low_array.id = 1 + node_array.id.max() + np.arange(amount)
        load_low_array.node = node_array.id
        load_low_array.status = 1
        load_high_array = self.grid.sym_load.__class__.zeros(amount)
        load_high_array.id = 1 + load_low_array.id.max() + np.arange(amount)
        load_high_array.node = node_array.id
        load_high_array.status = 1

        # power consumption in Watt
        load_low_array.p_specified = np.round(self.rng.normal(200_000, 150_000, amount))
        load_low_array.q_specified = np.round(self.rng.normal(20_000, 15_000, amount))
        load_high_array.p_specified = np.round(self.rng.normal(-100_000, 350_000, amount))
        load_high_array.q_specified = np.round(self.rng.normal(-5_000, 35_000, amount))

        return node_array, load_low_array, load_high_array



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Sensor data types
Based on the sensors defined in power grid model:
https://power-grid-model.readthedocs.io/en/v1.9.13/user_manual/components.html#sensor
"""

from typing import Annotated, Literal, TypeVar

import numpy as np
from numpy.typing import NDArray

from power_grid_model_ds._core.model.dtypes.id import Id

# define structural arrays with 3 values for 3-phase variables
# based on https://stackoverflow.com/a/72585748
_DT = TypeVar("_DT", bound=np.generic)
NDArray3 = Annotated[NDArray[_DT], Literal[3]]


class Sensor(Id):
    """Base class for sensor data type"""

    measured_object: NDArray[np.int32]


class GenericPowerSensor(Sensor):
    """Base class for power sensor data type"""

    measured_terminal_type: NDArray[np.int32]
    power_sigma: NDArray[np.float64]  # std of total power


class SymPowerSensor(GenericPowerSensor):
    """SymPowerSensor data type"""

    p_measured: NDArray[np.float64]  # measured active power
    q_measured: NDArray[np.float64]  # measured reactive power
    p_sigma: NDArray[np.float64]  # std of active power
    q_sigma: NDArray[np.float64]  # std of reactive power


class GenericVoltageSensor(Sensor):
    """Base class for voltage sensor data type"""


class SymVoltageSensor(GenericVoltageSensor):
    """SymVoltageSensor data type"""

    u_sigma: NDArray[np.float64]  # std of voltage
    u_measured: NDArray[np.float64]  # measured voltage
    u_angle_measured: NDArray[np.float64]  # measured phase


class AsymVoltageSensor(GenericVoltageSensor):
    """AsymVoltageSensor data type"""

    u_sigma: NDArray3[np.float64]  # std of 3 voltages
    u_measured: NDArray3[np.float64]  # measured 3 voltages
    u_angle_measured: NDArray3[np.float64]  # measured 3 phases



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Base data types"""

from typing import Any

import numpy as np
from numpy.typing import NDArray

from power_grid_model_ds._core.model.constants import empty


class Id:
    """Base dtype for id arrays"""

    _defaults: dict[str, Any] = {"id": empty}
    id: NDArray[np.int32]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Power flow functions and classes"""

from typing import Dict, Optional

import numpy as np
from numpy.typing import NDArray
from power_grid_model import CalculationMethod, PowerGridModel, initialize_array

from power_grid_model_ds._core.model.grids.base import Grid

PGM_ARRAYS = [
    "node",
    "line",
    "link",
    "transformer",
    "three_winding_transformer",
    "sym_load",
    "sym_gen",
    "source",
    "transformer_tap_regulator",
    "sym_power_sensor",
    "sym_voltage_sensor",
    "asym_voltage_sensor",
]


class PGMCoreException(Exception):
    """Raised when there is an error in running the power grid model"""


class PowerGridModelInterface:
    """Interface between the Grid and the PowerGridModel (pgm).

    - Can convert grid data to pgm input
    - Can calculate power flow
    - Can do batch calculations using pgm
    - Can update grid with output from power flow
    """

    def __init__(
        self,
        grid: Optional[Grid] = None,
        input_data: Optional[Dict] = None,
        system_frequency: float = 50.0,
    ):
        self.grid = grid or Grid.empty()
        self.system_frequency = system_frequency

        self.input_data = input_data or {}
        self.output_data: dict[str, NDArray] = {}
        self.model: Optional[PowerGridModel] = None

    def create_input_from_grid(self):
        """
        Create input for the PowerGridModel
        """
        for array_name in PGM_ARRAYS:
            pgm_array = self._create_power_grid_array(array_name=array_name)
            self.input_data[array_name] = pgm_array
        return self.input_data

    def create_grid_from_input_data(self, check_ids: bool = True) -> Grid:
        """
        Create Grid object from PowerGridModel input.
        Note that for some arrays, not all fields are available in the PowerGridModel input.
        In this case, the default values are used.

        Args:
            check_ids: if True, check if the ids are unique

        Returns a Grid object with the arrays filled with the PowerGridModel input.
        """
        for pgm_name in PGM_ARRAYS:
            if pgm_name in self.input_data:
                pgm_ds_array_class = getattr(self.grid, pgm_name).__class__
                pgm_ds_array = pgm_ds_array_class(self.input_data[pgm_name])
                self.grid.append(pgm_ds_array, check_max_id=False)
        if check_ids:
            self.grid.check_ids()
        return self.grid

    def calculate_power_flow(
        self,
        calculation_method: CalculationMethod = CalculationMethod.newton_raphson,
        update_data: Optional[Dict] = None,
        **kwargs,
    ):
        """Initialize the PowerGridModel and calculate power flow over input data.

        If input data is not available, self.create_input_from_grid() will be called to create it.

        Returns output of the power flow calculation (also stored in self.output_data)
        """
        self.model = self.model or self._setup_model()

        self.output_data = self.model.calculate_power_flow(
            calculation_method=calculation_method, update_data=update_data, **kwargs
        )
        return self.output_data

    def _create_power_grid_array(self, array_name: str) -> np.ndarray:
        """Create power grid model array"""
        internal_array = getattr(self.grid, array_name)
        pgm_array = initialize_array("input", array_name, internal_array.size)
        fields = self._match_dtypes(pgm_array.dtype, internal_array.dtype)
        pgm_array[fields] = internal_array.data[fields]
        return pgm_array

    def update_model(self, update_data: Dict):
        """
        Updates the power-grid-model using update_data, this allows for batch calculations

        Example:
            Example of update_data creation:

            >>> update_sym_load = initialize_array('update', 'sym_load', 2)
            >>> update_sym_load['id'] = [4, 7]  # same ID
            >>> update_sym_load['p_specified'] = [30e6, 15e6]  # change active power
            >>> # leave reactive power the same, no need to specify
            >>>
            >>> update_line = initialize_array('update', 'line', 1)
            >>> update_line['id'] = [3]  # change line ID 3
            >>> update_line['from_status'] = [0]  # switch off at from side
            >>> # leave to-side swichint status the same, no need to specify
            >>>
            >>> update_data = {
            >>>    'sym_load': update_sym_load,
            >>>    'line': update_line
            >>> }


        """
        self.model = self.model or self._setup_model()
        self.model.update(update_data=update_data)

    def update_grid(self) -> None:
        """
        Fills the output values in the grid for the values that are present
        """
        if not self.output_data:
            raise PGMCoreException("Can not update grid without output_data")
        for array_name in PGM_ARRAYS:
            if array_name in self.output_data.keys():
                internal_array = getattr(self.grid, array_name)
                pgm_output_array = self.output_data[array_name]
                fields = self._match_dtypes(pgm_output_array.dtype, internal_array.dtype)
                internal_array[fields] = pgm_output_array[fields]

    @staticmethod
    def _match_dtypes(first_dtype: np.dtype, second_dtype: np.dtype):
        return list(set(first_dtype.names).intersection(set(second_dtype.names)))  # type: ignore[arg-type]

    def _setup_model(self):
        self.input_data = self.input_data or self.create_input_from_grid()
        self.model = PowerGridModel(self.input_data, system_frequency=self.system_frequency)
        return self.model



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from power_grid_model_ds._core.data_source.generator.arrays.line import LineGenerator
from power_grid_model_ds._core.data_source.generator.arrays.node import NodeGenerator
from power_grid_model_ds._core.data_source.generator.arrays.source import SourceGenerator
from power_grid_model_ds._core.data_source.generator.arrays.transformer import TransformerGenerator
from power_grid_model_ds._core.data_source.generator.grid_generators import RadialGridGenerator

__all__ = ["RadialGridGenerator", "NodeGenerator", "LineGenerator", "TransformerGenerator", "SourceGenerator"]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Generator for LineArray"""

import numpy as np

from power_grid_model_ds._core.data_source.generator.arrays.base import BaseGenerator
from power_grid_model_ds._core.model.arrays import TransformerArray


class TransformerGenerator(BaseGenerator):
    """Generator for tranformer elements in the grid"""

    def run(self, amount: int) -> TransformerArray:
        """Generate transformers"""

        # Create transformers from 10kV to 3kV
        from_mask = self.grid.node.u_rated == 10_500
        from_nodes = self.rng.choice(self.grid.node.id[from_mask], amount, replace=True)
        to_mask = self.grid.node.u_rated == 3_000
        to_nodes = self.rng.choice(self.grid.node.id[to_mask], amount, replace=False)
        transformer_array = self.grid.transformer.__class__.zeros(amount)
        transformer_array.id = 1 + self.grid.max_id + np.arange(amount)
        transformer_array.from_node = from_nodes
        transformer_array.to_node = to_nodes
        transformer_array.from_status = [1] * amount
        transformer_array.to_status = [1] * amount
        transformer_array.u1 = [10_500] * amount
        transformer_array.u2 = [3_000] * amount
        transformer_array.sn = [30e6] * amount
        transformer_array.clock = [12] * amount
        transformer_array.uk = [0.203] * amount
        transformer_array.pk = [100e3] * amount

        return transformer_array



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import logging
from typing import Generator

import rustworkx as rx
from rustworkx import NoEdgeBetweenNodes
from rustworkx.visit import BFSVisitor, PruneSearch, StopSearch

from power_grid_model_ds._core.model.graphs.errors import MissingBranchError, MissingNodeError, NoPathBetweenNodes
from power_grid_model_ds._core.model.graphs.models._rustworkx_search import find_fundamental_cycles_rustworkx
from power_grid_model_ds._core.model.graphs.models.base import BaseGraphModel

_logger = logging.getLogger(__name__)


class RustworkxGraphModel(BaseGraphModel):
    """A wrapper around the graph from the 'rustworkx' package"""

    def __init__(self, active_only=False) -> None:
        super().__init__(active_only=active_only)
        self._graph: rx.PyGraph = rx.PyGraph()
        self._internal_to_external: dict[int, int] = {}
        self._external_to_internal: dict[int, int] = {}

    @property
    def nr_nodes(self):
        return self._graph.num_nodes()

    @property
    def nr_branches(self):
        return self._graph.num_edges()

    @property
    def external_ids(self) -> list[int]:
        return list(self._external_to_internal.keys())

    # pylint: disable=duplicate-code
    def external_to_internal(self, ext_node_id: int):
        try:
            return self._external_to_internal[ext_node_id]
        except KeyError as error:
            raise MissingNodeError(f"External node id '{ext_node_id}' does NOT exist!") from error

    def internal_to_external(self, int_node_id: int):
        return self._internal_to_external[int_node_id]

    def _add_node(self, ext_node_id: int):
        graph_node_id = self._graph.add_node(ext_node_id)
        self._external_to_internal[ext_node_id] = graph_node_id
        self._internal_to_external[graph_node_id] = ext_node_id

    def _delete_node(self, node_id: int):
        self._graph.remove_node(node_id)
        external_node_id = self._internal_to_external.pop(node_id)
        self._external_to_internal.pop(external_node_id)

    def _has_branch(self, from_node_id: int, to_node_id: int) -> bool:
        return self._graph.has_edge(from_node_id, to_node_id)

    def _has_node(self, node_id: int) -> bool:
        return self._graph.has_node(node_id)

    def _add_branch(self, from_node_id: int, to_node_id: int):
        self._graph.add_edge(from_node_id, to_node_id, None)

    def _delete_branch(self, from_node_id: int, to_node_id: int) -> None:
        try:
            self._graph.remove_edge(from_node_id, to_node_id)
        except NoEdgeBetweenNodes as error:
            raise MissingBranchError(f"No edge between (internal) nodes {from_node_id} and {to_node_id}") from error

    def _get_shortest_path(self, source: int, target: int) -> tuple[list[int], int]:
        path_mapping = rx.dijkstra_shortest_paths(self._graph, source, target)

        if target not in path_mapping:
            raise NoPathBetweenNodes(f"No path between internal nodes {source} and {target}")

        path_nodes = list(path_mapping[target])
        return path_nodes, len(path_nodes) - 1

    def _get_all_paths(self, source: int, target: int) -> list[list[int]]:
        return list(rx.all_simple_paths(self._graph, source, target))

    def _get_components(self, substation_nodes: list[int]) -> list[list[int]]:
        no_os_graph = self._graph.copy()
        for os_node in substation_nodes:
            no_os_graph.remove_node(os_node)
        components = rx.connected_components(no_os_graph)
        return [list(component) for component in components]

    def _get_connected(self, node_id: int, nodes_to_ignore: list[int], inclusive: bool = False) -> list[int]:
        visitor = _NodeVisitor(nodes_to_ignore)
        rx.bfs_search(self._graph, [node_id], visitor)
        connected_nodes = visitor.nodes
        if not inclusive:
            connected_nodes.remove(node_id)

        return connected_nodes

    def _in_branches(self, int_node_id: int) -> Generator[tuple[int, int], None, None]:
        return ((source, target) for source, target, _ in self._graph.in_edges(int_node_id))

    def _find_first_connected(self, node_id: int, candidate_node_ids: list[int]) -> int:
        visitor = _NodeFinder(candidate_nodes=candidate_node_ids)
        rx.bfs_search(self._graph, [node_id], visitor)
        if visitor.found_node is None:
            raise MissingNodeError(f"node {node_id} is not connected to any of the candidate nodes")
        return visitor.found_node

    def _find_fundamental_cycles(self) -> list[list[int]]:
        """Find all fundamental cycles in the graph using Rustworkx.

        Returns:
            list[list[int]]: A list of cycles, each cycle is a list of node IDs.
        """
        return find_fundamental_cycles_rustworkx(self._graph)

    def _all_branches(self) -> Generator[tuple[int, int], None, None]:
        return ((source, target) for source, target in self._graph.edge_list())


class _NodeVisitor(BFSVisitor):
    def __init__(self, nodes_to_ignore: list[int]):
        self.nodes_to_ignore = nodes_to_ignore
        self.nodes: list[int] = []

    def discover_vertex(self, v):
        if v in self.nodes_to_ignore:
            raise PruneSearch
        self.nodes.append(v)


class _NodeFinder(BFSVisitor):
    """Visitor that stops the search when a candidate node is found"""

    def __init__(self, candidate_nodes: list[int]):
        self.candidate_nodes = candidate_nodes
        self.found_node: int | None = None

    def discover_vertex(self, v):
        if v in self.candidate_nodes:
            self.found_node = v
            raise StopSearch
